$TITLE A multistage SP formulation for an international portfolio management model

$ONTEXT
*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*
* Hercules Vladimirou, Nikolas Topaloglou, Stavros A. Zenios
* HERMES Center of Excellence on Computational Finance and Economics
* School of Economics and Management
* University of Cyprus
* 75 Kallipoleos Street
* P.O.Box 20537, CY-1678 Nicosia, Cyprus
* October 2001
*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*
$OFFTEXT

* define runtime control parameters
$OFFSYMXREF
$OFFSYMLIST
OPTION LIMCOL=0;
OPTION LIMROW=0;
OPTION SOLPRINT=OFF;
OPTION RESLIM=1000000;
OPTION ITERLIM=100000;
* select optimization solver
* OPTION LP=OSL;

$offlisting;

$eolcom #

* redefine runtime control parameters depending
* on required level of output detail
$set DISPL 0              # 1 for detailed output (set definitions, etc), 2 if increased model response also required
$if NOT "%DISPL%" == "2" $goto CONT1
OPTION LIMROW = 20;
OPTION LIMCOL = 5;
OPTION SOLPRINT=ON;

$label CONT1

* define the set of countries, the base country and the foreign countries

SET C countries (markets) / USA, UK, GR, JP /;
ALIAS(C,CC);
SET base(C) the base country (always the first one in the country list);
base(C)=no;
base(C)$(ord(C) eq 1) = yes;
SET foreign(c);
foreign(C)=(not base(C));

* define the set of asset classes and determine the admissible investment securities 
* (i.e., available asset classes in each country/currency

SET CLASS the investment asset classes / Stk, Bnd1, Bnd3, Bnd7 /;
SET securities(C,CLASS) the set of admissible investment securities;
securities(C,CLASS)=yes;

$if "%DISPL%" == "0" $goto CONT2
display C, base, foreign;
display CLASS, securities;

$label CONT2

$INCLUDE tree2.dat
$INCLUDE Assets2.dat
$INCLUDE Probabilities2.inc
$INCLUDE initial.dat

* determine the successor relationship between node pairs
* as the reverse of the predecessor relationship
*/ successor(n,nn)=yes  if node n is a successor of node nn
* may be redundant/unnecessary (now used in determination of root node)
set successor(N,NN);
successor(N,NN)=predecessor(NN,N);

* determine the leaf nodes of the scenario tree
* a node is a leaf if it is not a predecessor to any other node
SET leaves(N);
leaves(N) = yes;
loop((N,NN), leaves(N)$predecessor(N,NN)=no;);

* determine the root node 
* the root is that node which is not a successor to any other node
SET root(N);
root(N)=yes;
loop((N,NN), root(N)$successor(N,NN)=no;);

* determine the intermediate nodes
* (it does not seem necessary/important to distinguish them, seem redundant)
SET intermediate(N);
intermediate(N)=((not leaves(N)) and (not root(N)));

PARAMETER level(N) the level of each node on the scenario tree;

* determine the level of each node on the scenario tree
* CAUTION: the logic below assumes that tree nodes are labeled (sequenced in node set N)
* in a breadth-first manner
* we need the node levels only if we will be doing stage-based traversals of the tree
level(root(N))=0;
loop((N,NN), level(NN)$predecessor(N,NN)=level(N)+1;);

* define an error_code for termination
SCALAR e_code  error for termination in data verification tests / 0 /;
SCALAR tmp     temporary parameter in data verification tests   / 0 /;

* check that there is only one root note
tmp=0;
loop(N, tmp$root(N)=tmp+1;);
if (((tmp > 1) or (tmp < 1)),
   e_code=1;
   display N, predecessor, successor, root, leaves, level, intermediate;
   display "ERROR code: ",e_code," Number of root nodes =",tmp;
   abort "Number of root nodes is not one";
   );

* verify that in the precedence specifications
* every node (except the root) has exactly one predecessor
loop(N$(not root(N)), 
    tmp=0;
    loop(NN, tmp$predecessor(NN,N)=tmp+1;);
    if (((tmp > 1) or (tmp < 1)),
       e_code=2;
       display "ERROR: Node ", N ," has ",tmp," predecessors (different than 1)";
       );
    );
if (e_code > 0,
   display N, predecessor, successor, root, leaves, level, intermediate;
   display "ERROR code: ",e_code;
   abort " Nodes with incorrect predecessors identified";
   );

* determine the depth of the scenario tree (number of stages)
SCALAR n_stages  depth of the scenario tree;
n_stages = smax(leaves(N), level(N)); 
* check that all leaf nodes have the same level (at the depth of the tree)
loop(N$leaves(N),
    if ((level(N) lt n_stages),
       e_code=3;
       tmp=level(N);
       display "ERROR: Leaf Node ",N ," has level ",tmp;
       display "while depth of the tree is ",n_stages; 
       );
    );
if ((e_code > 0),
   display  e_code, N, predecessor, successor, root, leaves, level, intermediate, n_stages;
   display "ERROR code: ",e_code;
   abort "Leaf nodes at levels other than the tree depth have been identified";
   );


* Determine the probabilities of all tree nodes
* CAUTION: we assume that probabilities were entered only for the leaf nodes
* for all other nodes, probabilities are consistently infered
Prob(N)$(not leaves(N)) = 0.0;
* first, we rescale the probabilities of leaves so that they sum to 1.0
tmp = sum(leaves(NN), Prob(NN));
Prob(N)$leaves(N) = Prob(N)/tmp;
for (tmp=n_stages-1 downto 0,
    loop(N,
        if ((level(N) eq tmp), Prob(N) = sum(NN$successor(NN,N), Prob(NN)););
        );
    );

* ensure that the root has probability 1.0
tmp=sum(root(N),  Prob(N));
if (((tmp < (1.0-1.0e-5)) or (tmp > (1.0+1.0e-5))),
   e_code = 4;
   display "ERROR code: ",e_code," Probability of root node other than 1.0 ",tmp
;
   abort "Root node must have probability 1.0";
   );


$if "%DISPL%" == "0" $goto CONT3
*display N, predecessor, successor, root, leaves, level, intermediate, Prob, n_stages;

$label CONT3


* enter exchange rate data
* CAUTION: exchange rates should always be quoted 
* in units of the base currency per one unit of the foreign currency
*
* ensure that the expected values of the spot rates
* at any subtree match the forward rate at the subtree root node
*
*Scalar tmp1  auxiliary parameter in scaling computations;
*loop((C,N)$((not leaves(N)) and (not base(C))),
*tmp=sum(successor(NN,N), Prob(NN)*spot_rate(NN,C));
*if (((tmp < forward_rate(N,C)-1.0e-5) or (tmp > forward_rate(N,C)+1.0e-5)),
*   tmp1 = forward_rate(N,C)-tmp;
*   loop(NN$successor(NN,N), spot_rate(NN,C) = spot_rate(NN,C) + tmp1;);
*   );
*);

* ensure that exchange rates and forward rates of the base currency are always 1.0
spot_rate(N,base(C))    = 1.0;
forward_rate(N,base(C)) = 1.0;

*initial(securities(base(c),class)) = 0;

$if "%DISPL%" == "0" $goto CONT4
display Prob, Price, spot_rate, forward_rate, initial, cashin;

$label CONT4

Variables
  tot_return      expected final return of portfolio 
  f_return(n)     portfolio return at leaf node n 
  buy(n,c,class)  units of asset classes bought at each tree node
  sell(n,c,class) units of asset classes sold   at each tree node
  hold(n,c,class) units of asset classes held   at each tree node
  spot_exchS(n,c) Amount of  currency C sold  in the spot markets per scenario tree node
  spot_exchB(n,c) Amount of currency C bought in the spot markets per scenario tree node
  f_exch(n,c)     currency exchange transactions in the forward markets per scenario tree node
* f_exch +ve means sale     of foreign currency (i.e., inflow in the base currency))
* f_exch -ve means purchase of foreign currency (i.e., outflow from the base currency))
  y(n)            auxiliary variables in CVaR definition
  z               Value-at-Risk variable (VaR)
  CVar            Conditional-Value-at-Risk variable at the end of the second stage;

Positive Variables buy, sell, hold, spot_exchS, spot_exchB,  y;


Parameter t_cost(class) proportional transaction cost for asset sales and purchases
          / Stk     0.0005
            Bnd1    0.0005
            Bnd3    0.0005
            Bnd7    0.0005 /;    

Scalar tex_cost    proportional transaction cost for money sales and purchases in spot market / 0.0001 /;
Scalar bt         tolerance level (threshold) in CVaR definition /0.95/;
Scalar Rho         Lower bound in expected Return               /0/;


Parameter initial_value  initial wealth;
initial_value= sum(root(N), sum(C, (cashin(N,C) + sum(securities(C,CLASS), initial(C,CLASS)*Price(N,C,CLASS)))*spot_rate(N,C)));

equations
  init_base(n,c)           cash balance contraints for base country at the root node
  cash_base(n,nn,c)        cash balance contraints for base country at each scenario tree node
  init_foreign(n,c)        cash balance contraints for foreign countries at the root node
  cash_foreign(n,nn,c)     cash balance contraints for foreign countries at each scenario tree node
  balance(n,nn,c,class)    asset balance constraints at each tree node
  init_balance(n,c,class)  asset balance constraint at the root
  flimitf(n,c)             Limit on forward currency sales
  flimitex(n,c)            Limit on forward currency sales
  SellLimit(n,nn,c,class)  Disallow short sales
  cvar_def                 definition of long term CVaR of portfolio returns
  FinalValue(n,nn,c)       Final Value of entire portfolio at leaf node n (in base currency units)
  EXP_RET                  Total Return  
  Errest                   Target expected return
  DevfVaR(n)               Splitting of Deviation from VaR;


* bound on initial sales
 sell.up(n,c,class)$(root(n) and securities(c,class)) = initial(c,class);
* bound on total expected final return (Parameterized Target Exp. Return)
*tot_return.lo = Rho;
 
 
FinalValue(predecessor(nn,n),base(c))$leaves(n).. 
 f_return(n) =E= (cashin(n,c) + sum(securities(c,class), hold(nn,c,class)*Price(n,c,class)) 
  + sum(foreign(cc), f_exch(nn,cc) + spot_rate(n,cc)*( cashin(n,cc) + sum(securities(cc,class), hold(nn,cc,class)*Price(n,cc,class)) - f_exch(nn,cc)*(1.0/forward_rate(nn,cc))) ))/initial_value - 1.0;

 
EXP_RET.. tot_return =E= sum(leaves(n), f_return(n)*Prob(n));

init_base(root(n),base(c))..  cashin(n,c) + sum(securities(c,class), sell(n,c,class)*Price(n,c,class)*(1-t_cost(class)) ) 
                        + sum(foreign(cc), spot_exchS(n,cc))*(1-tex_cost) =E= sum(securities(c,class), buy(n,c,class)*Price(n,c,class)*(1+t_cost(class)) ) + sum(foreign(cc), spot_exchB(n,cc))*(1+tex_cost);

cash_base(predecessor(nn,n),base(c))$(not leaves(n))..  cashin(n,c) + sum(securities(c,class), sell(n,c,class)*Price(n,c,class)*(1-t_cost(class)))
                        + sum(foreign(cc), spot_exchS(n,cc)*(1-tex_cost) + f_exch(nn,cc))
                         =E= sum(securities(c,class), buy(n,c,class)*Price(n,c,class)*(1+t_cost(class))) + sum(foreign(CC), spot_exchB(n,cc))*(1+tex_cost);


init_foreign(root(n),foreign(c))..  cashin(n,c) + sum(securities(c,class), sell(n,c,class)*Price(n,c,class)*(1-t_cost(class)))  + (1.0/spot_rate(n,c))*spot_exchB(n,c)*(1-tex_cost) 
                         =E= sum(securities(c,class), buy(n,c,class)*Price(n,c,class)*(1+t_cost(class))) + (1/spot_rate(n,c))*spot_exchS(n,c)*(1+tex_cost);

cash_foreign(predecessor(nn,n),foreign(c))$(not leaves(n))..  cashin(n,c) + sum(securities(c,class), sell(n,c,class)*Price(n,c,class)*(1-t_cost(class))) + (1.0/spot_rate(n,c))*spot_exchB(n,c)*(1-tex_cost) 
                         =E= sum(securities(c,class), buy(n,c,class)*Price(n,c,class)*(1+t_cost(class)))
                        + (1.0/spot_rate(n,c))*spot_exchS(n,c)*(1+tex_cost) + (1.0/forward_rate(nn,c))*f_exch(nn,c);

init_balance(root(n),securities(c,class))..
  hold(n,c,class) =E= initial(c,class) + buy(n,c,class) - sell(n,c,class);

balance(predecessor(n,nn),securities(c,class))$((not leaves(nn)) and (not root(nn)))..
  hold(nn,c,class) =E= hold(n,c,class) + buy(nn,c,class) - sell(nn,c,class);

SellLimit(predecessor(nn,n),c,class)$((not leaves(n)) and (not root(n)))..  sell(n,c,class) =L= hold(nn,c,class);

* In definition of CVaR the loss function is defined as -(return) (i.e., -ve return) 
DevfVaR(n)$leaves(n)..    y(n) =G= z - f_return(n);
 
cvar_def..                CVar =E= z - (1.0/(1.0-bt))*sum(leaves(n), Prob(n)*y(n)); 

flimitf(n,c)$((not leaves(n)) and (not base(C)))..    f_exch(n,c) =L= sum(class, hold(n,c,class)*price(n,c,class))*spot_rate(n,c);

flimitex(nn,c)$((not leaves(nn)) and (not base(C)))..    f_exch(nn,c) =L= sum((n,class)$(predecessor(nn,n)), 
 Prob(n)*hold(nn,c,class)*price(n,c,class)*spot_rate(n,c))/Prob(nn);

Errest..        tot_return =G= Rho;

model portfolio1 / init_base, cash_base, init_foreign, cash_foreign, balance, init_balance, SellLimit, cvar_def, FinalValue, EXP_RET, DevfVaR, flimitex /;

model portfolio2 / init_base, cash_base, init_foreign, cash_foreign, balance, init_balance, SellLimit, FinalValue, EXP_RET, flimitex /;

model portfolio3 / Errest,  init_base, cash_base, init_foreign, cash_foreign, balance, init_balance, SellLimit, cvar_def, FinalValue, EXP_RET, DevfVaR, flimitex /;

* We find ten optimal points in order to construct the efficient frontiers
* We parameterize the expected return while we
* optimize the Conditional-Value-At-Risk


SET ITER /1*11/;

PARAMETERS
SZEROs
INCRs
SZEROn
INCRn;

* Define the output files

file cvar2, stcvar2, invest2;
cvar2.ap=1; stcvar2.ap=1; invest2.ap=1;

* We use selective currency hedging via forward contracts

* We find the minimum risk point on the efficient frontiers

Solve portfolio1 using lp maximizing CVaR;
SZEROs = tot_return.l;

* We find the maximum return point on the efficient frontiers

solve portfolio2 using lp maximizing tot_return;

* devide the distance between the minimum and maximum return in 10 equal intervals

INCRs = (tot_return.l-SZEROs)/10;

put cvar2;
put 'With Selective Hedging'
put /;

put stcvar2;
put 'With Selective Hedging'
put /;

put invest2;
put 'With Selective Hedging'
put /;

LOOP(ITER,
Rho=SZEROs+(ORD(ITER)-1)*INCRs- 0.0001;


Solve portfolio3 using lp maximizing CVaR;

* output file cvar2 includes the optimal Var, CVaR and Expected return of the optimal portfolios

put cvar2;
cvar2.nd=6;
put 'var, cvar, expected return', z.l, CVar.l, tot_return.l  /;

* output file invest2 includes the positions in each asset of each optimal portfolio

put invest2;
invest2.nd=6;
put 'Units of assets invested in first period  '/;
put '/ '/;
loop((C,CLASS), put  C.tl , '.' CLASS.tl @20, hold.L( 'n150',C,CLASS)  /);
put '/ '/;
put 'transactions in spot market (sales and purchases)  '/;
loop(C, put @3,  C.tl @15 spot_exchS.l('n150',C) @30 spot_exchB.l('n150',C)   /) ;
put '/ '/;
put 'transactions in forward market  '/;
loop(C, put @3,  C.tl @15 f_exch.L('n150',C) /);

* output file STcvar1 includes information about the Status of the model and the solver


put STcvar2;
put 'SOLVER STATUS', portfolio3.solvestat /;
put 'MODEL STATUS', portfolio3.modelstat /;
put 'Number of single variables generated', portfolio3.numvar /;
put  'Number of single  generated', portfolio3.numequ /;
put 'Number of non-zero entries in the coefficient matrix', portfolio3.numnz /;
put /;
);

* We use No currency hedging via forward contracts

Rho = 0;
f_exch.fx(n,c)=0;

Solve portfolio1 using lp maximizing CVaR;
SZEROn = tot_return.l;

solve portfolio2 using lp maximizing tot_return;

INCRn = (tot_return.l-SZEROn)/10;

put cvar2;
put 'Without  Hedging'
put /;

put stcvar2;
put 'Without  Hedging'
put /;

put invest2;
put 'Without  Hedging'
put /;

LOOP(ITER,
Rho=SZEROn+(ORD(ITER)-1)*INCRn-0.0001;

Solve portfolio3 using lp maximizing CVaR;

put cvar2;
cvar2.nd=6;
put 'var, cvar, expected return', z.l, CVar.l, tot_return.l  /;

put invest2;
invest2.nd=6;
put 'Units of assets invested in first period  '/;
put '/ '/;
loop((C,CLASS), put  C.tl , '.' CLASS.tl @20, hold.L( 'n150',C,CLASS)  /);
put '/ '/;
put 'transactions in spot market (sales and purchases)  '/;
loop(C, put @3,  C.tl @15 spot_exchS.l('n150',C) @30 spot_exchB.l('n150',C)   /) ;
put '/ '/;
put 'transactions in forward market  '/;
loop(C, put @3,  C.tl @15 f_exch.L('n150',C) /);

put STcvar2;
put 'SOLVER STATUS', portfolio3.solvestat /;
put 'MODEL STATUS', portfolio3.modelstat /;
put 'Number of single variables generated', portfolio3.numvar /;
put  'Number of single  generated', portfolio3.numequ /;
put 'Number of non-zero entries in the coefficient matrix', portfolio3.numnz /;
put /;
);
